/* This file is automatically generated; do not edit.     */
/* Add the files to be included into Makefile.am instead. */

#include <ql/experimental/variancegamma/analyticvariancegammaengine.hpp>
#include <ql/experimental/variancegamma/fftengine.hpp>
#include <ql/experimental/variancegamma/fftvanillaengine.hpp>
#include <ql/experimental/variancegamma/fftvariancegammaengine.hpp>
#include <ql/experimental/variancegamma/variancegammamodel.hpp>
#include <ql/experimental/variancegamma/variancegammaprocess.hpp>

