#projectname
{
    font-size: 16px;
    padding: 1em;
}

#projectbrief
{
    font-family: 'Merriweather Sans', serif;
    font-size: 120%;
    margin-top: 2px;
    margin-bottom: 0;
}

#projectnumber
{
    font-family: 'Merriweather Sans', serif;
    font-size: 100%;
    margin-top: 1em;
    margin-bottom: 0;
}

