/* This file is automatically generated; do not edit.     */
/* Add the files to be included into Makefile.am instead. */

#include <ql/indexes/ibor/aonia.hpp>
#include <ql/indexes/ibor/audlibor.hpp>
#include <ql/indexes/ibor/bbsw.hpp>
#include <ql/indexes/ibor/bibor.hpp>
#include <ql/indexes/ibor/bkbm.hpp>
#include <ql/indexes/ibor/cadlibor.hpp>
#include <ql/indexes/ibor/cdor.hpp>
#include <ql/indexes/ibor/chflibor.hpp>
#include <ql/indexes/ibor/dkklibor.hpp>
#include <ql/indexes/ibor/eonia.hpp>
#include <ql/indexes/ibor/euribor.hpp>
#include <ql/indexes/ibor/eurlibor.hpp>
#include <ql/indexes/ibor/fedfunds.hpp>
#include <ql/indexes/ibor/gbplibor.hpp>
#include <ql/indexes/ibor/jibar.hpp>
#include <ql/indexes/ibor/jpylibor.hpp>
#include <ql/indexes/ibor/libor.hpp>
#include <ql/indexes/ibor/mosprime.hpp>
#include <ql/indexes/ibor/nzdlibor.hpp>
#include <ql/indexes/ibor/nzocr.hpp>
#include <ql/indexes/ibor/pribor.hpp>
#include <ql/indexes/ibor/robor.hpp>
#include <ql/indexes/ibor/seklibor.hpp>
#include <ql/indexes/ibor/shibor.hpp>
#include <ql/indexes/ibor/sofr.hpp>
#include <ql/indexes/ibor/sonia.hpp>
#include <ql/indexes/ibor/tibor.hpp>
#include <ql/indexes/ibor/thbfix.hpp>
#include <ql/indexes/ibor/trlibor.hpp>
#include <ql/indexes/ibor/usdlibor.hpp>
#include <ql/indexes/ibor/wibor.hpp>
#include <ql/indexes/ibor/zibor.hpp>

